package fr.insa.clavardator.client.config;

import fr.insa.clavardator.client.server.PresenceType;
import org.json.JSONException;
import org.json.JSONObject;

@SuppressWarnings("FieldCanBeLocal")
public class Config {

	private ServerConfig serverConfig;
	private final String SERVER_KEY = "serveur";

	public Config() {
		serverConfig = new ServerConfig();
	}

	public Config(JSONObject obj) {
		this();
		if (obj.has(SERVER_KEY)) {
			this.serverConfig = new ServerConfig(obj.getJSONObject(SERVER_KEY));
		}
	}

	public ServerConfig getServerConfig() {
		return serverConfig;
	}

	@SuppressWarnings("FieldCanBeLocal")
	public static class ServerConfig {

		private JSONObject obj;

		private boolean enabled;
		private String uri;
		private PresenceType type;
		private int presencePort;
		private int proxyPort;

		private final String ENABLED_KEY = "actif";
		private final String URI_KEY = "uri";
		private final String TYPE_KEY = "type";
		private final String PORTS_KEY = "ports";
		private final String PORT_PRESENCE_KEY = "presence";
		private final String PORT_PROXY_KEY = "proxy";

		/**
		 * Basic constructor setting the default server configuration
		 */
		public ServerConfig() {
			enabled = false;
			uri = "";
			type = PresenceType.INSA;
		}

		/**
		 * Tries to read the given JSON object to extract custom user configuration
		 *
		 * @param obj THe JSON object to read config from
		 */
		public ServerConfig(JSONObject obj) {
			this();
			this.obj = obj;
			readServerEnabled();
			if (enabled) {
				readServerUri();
				if (uri.isBlank()) {
					enabled = false;
				} else {
					readServerType();
					readServerPorts();
				}
			}
		}

		/**
		 * Reads if the server is enabled.
		 * Uses false by default.
		 */
		private void readServerEnabled() {
			try {
				enabled = obj.getInt(ENABLED_KEY) == 1;
			} catch (JSONException e) {
				enabled = false;
			}
		}

		/**
		 * Reads the server URI.
		 * Uses the empty string as default
		 */
		private void readServerUri() {
			try {
				uri = obj.getString(URI_KEY);
			} catch (JSONException e) {
				uri = "";
			}
		}

		/**
		 * Reads the server type.
		 * Uses INSA as default.
		 */
		private void readServerType() {
			try {
				type = obj.getEnum(PresenceType.class, TYPE_KEY);
			} catch (JSONException e) {
				type = PresenceType.INSA;
			}
		}

		private void readServerPorts() {
			try {
				JSONObject portsObj = obj.getJSONObject(PORTS_KEY);
				presencePort = portsObj.getInt(PORT_PRESENCE_KEY);
				proxyPort = portsObj.getInt(PORT_PROXY_KEY);
			} catch (JSONException e) {
				presencePort = 30000; // TODO set default ports
				proxyPort = 300001;
			}
		}

		/**
		 * Checks if the presence server is enabled.
		 * If this returns false, all other data returned by this object can be ignored.
		 *
		 * @return True if enabled, false otherwise.
		 */
		public boolean isEnabled() {
			return enabled;
		}

		public String getUri() {
			return uri;
		}

		public PresenceType getType() {
			return type;
		}

		public int getPresencePort() {
			return presencePort;
		}

		public int getProxyPort() {
			return proxyPort;
		}
	}
}
